
#define BTRFS_MAGIC "_BHRfS_M"
#define le64_to_cpu(x) ((__u64) (x))
#define BTRFS_MAX_CHUNK_ENTRIES 256
#define BTRFS_MAX_LEAF_SIZE 4096
#define BTRFS_MAX_LEVEL 8
#define BTRFS_UUID_SIZE 16
#define BTRFS_CSUM_SIZE 32
#define BTRFS_FSID_SIZE 16
#define BTRFS_SYSTEM_CHUNK_ARRAY_SIZE 2048
#define BTRFS_LABEL_SIZE 256
#define BTRFS_SUPER_MIRROR_MAX  3
#define BTRFS_SUPER_INFO_OFFSET (64 * 1024)
#define BTRFS_SUPER_FLAG_METADUMP       (1ULL << 33)

#define BTRFS_INODE_ITEM_KEY    1
#define BTRFS_DIR_ITEM_KEY      84
#define BTRFS_DIR_INDEX_KEY     96
#define BTRFS_EXTENT_DATA_KEY   108
#define BTRFS_ROOT_ITEM_KEY     132
#define BTRFS_ROOT_REF_KEY      156
#define BTRFS_CHUNK_ITEM_KEY    228


#define BTRFS_FS_TREE_OBJECTID 5ULL
#define BTRFS_FIRST_FREE_OBJECTID 256ULL
#define BTRFS_FIRST_CHUNK_TREE_OBJECTID 256ULL

#define BTRFS_FILE_EXTENT_INLINE 0
#define MAX_LINK_COUNT           5
#define PATH_MAX                 1024


//<TODO> temporary - copy in the original code from grub somewhere. not done
#define SECTOR_BITS 9 // sector size is 512 bytes which is 2^9



#define __le64 long long
#define __le32 int
#define __le16 short
#define u8 unsigned char
#define u16 unsigned short
#define u32 unsigned int
#define u64 unsigned long long


enum dirent_type {
    DT_UNKNOWN  =  0,
    DT_FIFO     =  1,
    DT_CHR      =  2,
    DT_DIR      =  4,
    DT_BLK      =  6,
    DT_REG      =  8,
    DT_LNK      = 10,
    DT_SOCK     = 12,
    DT_WHT      = 14,
};



struct extent_buffer{
	u64 start;
	u64 dev_bytenr;
	u32 len;
	char * data;
};

struct btrfs_timespec {
       __le64 sec;
       __le32 nsec;
} __attribute__ ((__packed__));


struct btrfs_inode_item {
        /* nfs style generation number */
        __le64 generation;
        /* transid that last touched this inode */
        __le64 transid;
        u64 size;
        __le64 nbytes;
        __le64 block_group;
        __le32 nlink;
        __le32 uid;
        __le32 gid;
        __le32 mode;
        __le64 rdev;
        __le64 flags;

        /* modification sequence number for NFS */
        __le64 sequence;
        /*
         * a little future expansion, for more than this we can
         * just grow the inode item and version it
         */
        __le64 reserved[4];
        struct btrfs_timespec atime;
        struct btrfs_timespec ctime;
        struct btrfs_timespec mtime;
        struct btrfs_timespec otime;
}__attribute__ ((__packed__));

struct btrfs_disk_key {
        __le64 objectid;
        u8 type;
        __le64 offset;
}__attribute__ ((__packed__));

struct btrfs_root_item{
	struct btrfs_inode_item inode;
        __le64 generation;
        __le64 root_dirid;
        __le64 bytenr;
        __le64 byte_limit;
        __le64 bytes_used;
        __le64 last_snapshot;
        __le64 flags;
        __le32 refs;
        struct btrfs_disk_key drop_progress;
        u8 drop_level;
        u8 level;
}__attribute__ ((__packed__));

struct btrfs_root{
	struct extent_buffer   node;
	char                   data[4096];
	struct btrfs_root_item root_item;
	u64 objectid;

	/* data allocations are done in sectorsize units */
	u32 sectorsize;
	/* node allocations are done in nodesize units */
	u32 nodesize;
	/* leaf allocations are done in leafsize units */
	u32 leafsize;
	/* leaf allocations are done in leafsize units */
	u32 stripesize;

	u32 type;
}__attribute__ ((__packed__));

/* this represents a divice in a chunk tree */
struct btrfs_dev_item {
       __le64 devid; /* internal device id */
       __le64 total_bytes; /* size of the device */
       __le64 bytes_used;
       __le32 io_align; /* optimal io alignment */
       __le32 io_width; /* optimal io width */
       __le32 sector_size; /* minimal io size */
       __le64 type; /* type and info about this device */
       __le64 generation; /* expected generation */
        __le64 start_offset; /* of the partition on a device */

       /* info for allocation decisions */
       __le32 dev_group;

        u8 seek_speed; /* 0-100 (100 is fastest) */
       u8 bandwidth;  /* 0-100 (100 is fastest) */

        u8 uuid[BTRFS_UUID_SIZE]; /* dev uuid generated by btrfs */
       u8 fsid[BTRFS_UUID_SIZE]; /* uuid of the host FS */
} __attribute__ ((__packed__));


struct btrfs_super_block{
	u8 csum[BTRFS_CSUM_SIZE];
	/* the first 4 fields must match struct btrfs_header */
	u8 fsid[BTRFS_FSID_SIZE];    /* FS specific uuid */
	u64 bytenr; /* this block number */
	__le64 flags;

	/* allowed to be different from the btrfs_header from here own down */
	__le64 magic;
	u64 generation;
	__le64 root;
	__le64 chunk_root;
	__le64 log_root;

	/* this will help find the new super based on the log root */
	__le64 log_root_transid;
	__le64 total_bytes;
	__le64 bytes_used;
	__le64 root_dir_objectid;
	__le64 num_devices;
	__le32 sectorsize;
	__le32 nodesize;
	__le32 leafsize;
	__le32 stripesize;
	__le32 sys_chunk_array_size;
	__le64 chunk_root_generation;
	__le64 compat_flags;
	__le64 compat_ro_flags;
	__le64 incompat_flags;
	__le16 csum_type;
	u8 root_level;
	u8 chunk_root_level;
	u8 log_root_level;
	struct btrfs_dev_item dev_item;

	char label[BTRFS_LABEL_SIZE];

	/* future expansion */
	__le64 reserved[32];
	u8 sys_chunk_array[BTRFS_SYSTEM_CHUNK_ARRAY_SIZE];
}__attribute__ ((__packed__));


struct btrfs_stripe {
        __le64 devid;
        __le64 offset;
        u8 dev_uuid[BTRFS_UUID_SIZE];
} __attribute__ ((__packed__));



struct btrfs_chunk {
        __le64 length;
        __le64 owner;
        __le64 stripe_len;
        __le64 type;
        __le32 io_align;
        __le32 io_width;
        __le32 sector_size;
        __le16 num_stripes;
        __le16 sub_stripes;
        struct btrfs_stripe stripe;
} __attribute__ ((__packed__));


/* store logical offset to physical offset mapping */
struct btrfs_chunk_map_item {
        u64 logical;
        u64 length;
        u64 devid;
        u64 physical;
}__attribute__ ((__packed__));

struct btrfs_chunk_map {
	struct btrfs_chunk_map_item *map;
        u32 map_length;
        u32 cur_length;
}__attribute__ ((__packed__));


struct btrfs_fs_info {
	u8 fsid[BTRFS_FSID_SIZE];
	struct btrfs_super_block sb_copy;
	struct btrfs_root chunk_root;
	//stores the logical bytenr of where the fs_root is stored
	u64 fs_root;
	u64 sb_transid;
	int fd;
	int sb_mirror;
	struct btrfs_chunk_map chunk_map;
}__attribute__ ((__packed__));

/* parent_inum for resolving "..", resolved_inum for getting the resolved
 * inode given an inode of a symbolic link. path is used for fetching the
 * resolved inode given the path instead of the inode num*/

struct btrfs_cache {
    __le64 inum;
    __le64 resolved_inum;
    __le64 parent_inum;
    char * path;
    struct btrfs_cache * next;
}__attribute__((__packed__));

/* Information about a "mounted" btrfs filesystem.  */
struct grub_btrfs_data
{
	struct btrfs_fs_info *fs_info;
	struct btrfs_inode_item inode;
	grub_disk_t disk; //pointer to a disk
	u64 ino;
	u64 offset;
        struct btrfs_cache * cache;
}__attribute__ ((__packed__));

struct btrfs_fs_cache
{
    struct btrfs_fs_info fs_info;
    struct btrfs_cache dentry_cache;
    int populated;
    unsigned long disk_id;
}__attribute__ ((__packed__));

struct btrfs_dir_item {
        struct btrfs_disk_key location;
        __le64 transid;
        __le16 data_len;
        __le16 name_len;
        u8 type;
} __attribute__ ((__packed__));

/* Item header for per-leaf lookup */
struct btrfs_item {
       struct btrfs_disk_key key;
       __le32 offset;
       __le32 size;
} __attribute__ ((__packed__));


/* remember how we get to a node/leaf */
struct btrfs_path {
        u64 offsets[BTRFS_MAX_LEVEL];
        int itemsnr[BTRFS_MAX_LEVEL];
        int slots[BTRFS_MAX_LEVEL];
        /* remember last slot's item and data */
        struct btrfs_item item;
        u8 data[BTRFS_MAX_LEAF_SIZE];
}__attribute__ ((__packed__));

struct btrfs_header {
        /* these first four must match the super block */
        u8 csum[BTRFS_CSUM_SIZE];
        u8 fsid[BTRFS_FSID_SIZE]; /* FS specific uuid */
        __le64 bytenr; /* which block this node is supposed to live in */
        __le64 flags;

        /* allowed to be different from the super from here on down */
        u8 chunk_tree_uuid[BTRFS_UUID_SIZE];
        __le64 generation;
        __le64 owner;
        __le32 nritems;
        u8 level;
} __attribute__ ((__packed__));

struct btrfs_leaf {
        struct btrfs_header header;
        struct btrfs_item items[];
} __attribute__ ((__packed__));


struct btrfs_key_ptr {
        struct btrfs_disk_key key;
        __le64 blockptr;
        __le64 generation;
} __attribute__ ((__packed__));

struct btrfs_node {
        struct btrfs_header header;
        struct btrfs_key_ptr ptrs[];
} __attribute__ ((__packed__));

struct btrfs_file_extent_item {
        __le64 generation;
        __le64 ram_bytes;
        u8 compression;
        u8 encryption;
        __le16 other_encoding; /* spare for later use */
        u8 type;
        u64 disk_bytenr;
        __le64 disk_num_bytes;
        __le64 offset;
        __le64 num_bytes;
} __attribute__ ((__packed__));

struct btrfs_root_ref {
        __le64 dirid;
        __le64 sequence;
        __le16 name_len;
} __attribute__ ((__packed__));
        


